#include <bits/stdc++.h>

using namespace std;

#ifdef TEST
#   define DEBUG(x) cout << (x) << endl
#else
#   define DEBUG(x)
#endif

struct node {
    int weight;
    node* parent;
    char edge;
    
    node(int w = 0) : weight(w), parent(NULL) {}
    node(node* l, node* r) : weight(l->weight + r->weight), parent(NULL) { 
        l->parent = r->parent = this;
        l->edge = '0';
        r->edge = '1';
    }
    
    bool operator<(const node& r) const { return weight < r.weight; }
};

string huffman_coding(const string& in_string) {
    // compute frequencies of chars
    vector<char> chars;
    unordered_map<char, int> freq;
    for(auto c : in_string) if(++freq[c] == 1) chars.push_back(c);

    // each char is a node whose weight is its frequency
    vector<node*> nodes;
    for(auto c : chars) nodes.push_back(new node(freq[c]));
    
    // iteratively merge the two less frequent nodes
    priority_queue<node*> q(nodes.begin(), nodes.end());
    while(q.size() > 1) {
        auto a = q.top();
        q.pop();
        auto b = q.top();
        q.pop();
        nodes.push_back(new node(b, a));
        q.push(nodes.back());
    }
    
    // compute wordcodes starting from leaves and reaching the root
    unordered_map<char, string> wordcode;
    for(int i = 0; i < chars.size(); i++) {
        stack<char> s;
        for(node* n = nodes[i]; n->parent != NULL; n = n->parent) s.push(n->edge);
        while(!s.empty()) { wordcode[chars[i]] += s.top(); s.pop(); }
    }

    // free dynamic memory    
    for(auto x : nodes) delete x;

    // assemble the answer
    string res;
    for(auto c : in_string) res += wordcode[c];
    return res;
}

int main() {
    cout << huffman_coding("AABACDACA") << endl;
    return 0;
}

