/*

Given a set of n integers, where n <= 40 and their sum is at most 10^30,
determine the maximum sum subset having sum less than or equal S.

*/

#include <bits/stdc++.h>

using namespace std;

#define N 24

void generate_sums(int array[], int size, int upper_bound, vector<int>& res) {
    res.push_back(0);
    unordered_set<int> seen = {0};
    for(int i = 0; i < size; i++) {
        for(int j = res.size() - 1; j >= 0; j--) {
            if(array[i] + res[j] < upper_bound && seen.insert(array[i] + res[j]).second) {
                res.push_back(array[i] + res[j]);
            }
        }
    }
}

int solve(int array[], int size, int S) {
    vector<int> all_sums;
    generate_sums(array, size, S + 1, all_sums);
    sort(all_sums.begin(), all_sums.end(), greater<int>());
    auto it = lower_bound(all_sums.begin(), all_sums.end(), S, greater<int>());
    return *it;
}

int solve_with_meet_in_the_middle(int array[], int size, int S) {
    // split the array and generate all sums of the two halves
    vector<int> all_sums_1, all_sums_2;
    generate_sums(array, size / 2, S + 1, all_sums_1);
    generate_sums(array + size / 2, (size / 2 + size % 2), S + 1, all_sums_2);

    // sort the two sets of sums with different criteria
    sort(all_sums_1.begin(), all_sums_1.end());
    sort(all_sums_2.begin(), all_sums_2.end(), greater<int>());
    
    // sum up the last elements, and remove one of the two
    int res = 0;
    while(!all_sums_1.empty() && !all_sums_2.empty()) {
        int s = all_sums_1.back() + all_sums_2.back();
        if(s > S) { all_sums_1.pop_back(); continue; }
        res = max(res, s);
        all_sums_2.pop_back();
    }
    return res;
}

int main() {
    int array[N];
    
    // generate the worst case (each element sets a different bit)
    for(int i = 0; i < N; i++) array[i] = 1 << i;
    
    {
        clock_t start_time = clock();
        int res = solve(array, N, 12345678);
        clock_t elapsed_time = clock() - start_time;
        cout << "res = " << res << "\ntime (s) = " << fixed << static_cast<double>(elapsed_time) / CLOCKS_PER_SEC << endl;
    }
    
    {
        clock_t start_time = clock();
        int res = solve_with_meet_in_the_middle(array, N, 12345678);
        clock_t elapsed_time = clock() - start_time;
        cout << "res = " << res << "\ntime (s) = " << fixed << static_cast<double>(elapsed_time) / CLOCKS_PER_SEC << endl;
    }
    
    
    return 0;
}
